use std::any::{Any, TypeId};
use std::collections::HashSet;
use std::rc::Rc;
use std::time::Duration;

pub use sdl2::keyboard::Keycode;

use sdl2::{
    event::Event,
    image::{self, LoadTexture, Sdl2ImageContext},
    pixels::Color,
    rect::Rect,
    render::{Canvas, Texture, TextureCreator},
    video::{Window, WindowBuildError, WindowContext},
    IntegerOrSdlError, Sdl, VideoSubsystem,
};

#[derive(Debug, Clone)]
pub struct Error(String);

impl std::fmt::Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

macro_rules! impl_from_T_for_Error {
    ($t:ty) => {
        impl From<$t> for Error {
            fn from(value: $t) -> Self {
                Self(value.to_string())
            }
        }
    };
    ($t:ty, $($ts:ty),+) => {
        impl From<$t> for Error {
            fn from(value: $t) -> Self {
                Self(value.to_string())
            }
        }
        impl_from_T_for_Error!($($ts),+);
    };
}

impl_from_T_for_Error!(String, WindowBuildError, IntegerOrSdlError, &str);

pub type Id = u64;

pub struct Entity(Id, Vec<Box<dyn Component>>);

pub trait Component {
    fn inner_type_id(&self) -> TypeId;
    fn as_any(&mut self) -> &mut dyn Any;
}

#[derive(Clone, Copy)]
pub struct Sprite(Id);

pub struct Context<'context, 'game>
where
    'game: 'context,
{
    id_counter: &'context mut Id,
    canvas: &'context mut Canvas<Window>,
    texture_creator: *const TextureCreator<WindowContext>,
    entities: &'context mut Vec<Entity>,
    systems: &'context mut Vec<Rc<dyn System>>,
    textures: &'context mut Vec<(Id, Texture<'game>)>,
    currently_pressed_keys: &'context HashSet<Keycode>,
}

impl<'context, 'game> Context<'context, 'game> {
    pub fn entities_with_component<T: 'static + Component>(&self) -> Vec<u64> {
        let entity_type_id = TypeId::of::<T>();
        self.entities
            .iter()
            .filter_map(|Entity(id, components)| {
                let contains_component = components
                    .iter()
                    .any(|entity| (*entity).inner_type_id() == entity_type_id);
                if contains_component {
                    Some(*id)
                } else {
                    None
                }
            })
            .collect()
    }

    pub fn entity_component<T: 'static + Component>(&mut self, entity_id: u64) -> &mut T {
        let entity_type_id = TypeId::of::<T>();
        let Entity(_id, components) = self
            .entities
            .iter_mut()
            .find(|Entity(id, _)| *id == entity_id)
            .unwrap();

        let component = components
            .iter_mut()
            .find_map(|entity| {
                let is_id = (*entity).inner_type_id() == entity_type_id;
                if is_id {
                    Some(entity.as_any().downcast_mut::<T>().unwrap())
                } else {
                    None
                }
            })
            .unwrap();
        component
    }

    pub fn load_sprite<P>(&mut self, path: P) -> Result<Sprite, Error>
    where
        P: AsRef<std::path::Path>,
    {
        let path = path.as_ref().to_path_buf();
        let texture: Texture<'game> = unsafe { (*self.texture_creator).load_texture(&path)? };
        let id = *self.id_counter;
        *self.id_counter += 1;

        self.textures.push((id, texture));
        Ok(Sprite(id))
    }

    pub fn draw_sprite(&mut self, sprite: &Sprite, x: i32, y: i32) -> Result<(), Error> {
        let texture = self
            .textures
            .iter()
            .find_map(|v| if v.0 == sprite.0 { Some(&v.1) } else { None })
            .ok_or_else(|| "invalid sprite id")?;
        self.canvas.copy(
            texture,
            None,
            Rect::new(
                x * 4,
                y * 4,
                texture.query().width * 4,
                texture.query().height * 4,
            ),
        )?;
        Ok(())
    }

    pub fn spawn(&mut self, components: Vec<Box<dyn Component>>) {
        let id = *self.id_counter;
        *self.id_counter += 1;
        self.entities.push(Entity(id, components));
    }

    pub fn add_system(&mut self, system: Rc<dyn System>) {
        system.on_add(self);
        self.systems.push(system)
    }

    pub fn key_pressed(&self, keycode: Keycode) -> bool {
        self.currently_pressed_keys.contains(&keycode)
    }
}

pub trait System {
    fn on_add(&self, _ctx: &mut Context) {}
    fn on_update(&self, _ctx: &mut Context) -> Result<(), Error> {
        Ok(())
    }
}

pub struct Game<'a> {
    id_counter: u64,
    sdl_context: Sdl,
    video_subsystem: VideoSubsystem,
    image_context: Sdl2ImageContext,
    canvas: Canvas<Window>,
    texture_creator: TextureCreator<WindowContext>,
    event_pump: sdl2::EventPump,
    entities: Vec<Entity>,
    components: Vec<(u64, Box<dyn Component>)>,
    systems: Vec<Rc<dyn System>>,
    textures: Vec<(Id, Texture<'a>)>,
    currently_pressed_keys: HashSet<Keycode>,
}

impl<'game> Game<'game> {
    pub fn new() -> Result<Self, Error> {
        let sdl_context = sdl2::init()?;
        let video_subsystem = sdl_context.video()?;
        let image_context = image::init(image::InitFlag::PNG)?;
        let window = video_subsystem
            .window("pvp-game-dilapidation", 1280, 720)
            .position_centered()
            .build()?;

        let mut canvas = window.into_canvas().build()?;
        let texture_creator = canvas.texture_creator();

        canvas.set_draw_color(Color::RGB(60, 180, 180));
        canvas.clear();
        canvas.present();
        let event_pump = sdl_context.event_pump()?;
        Ok(Self {
            id_counter: 0,
            sdl_context,
            video_subsystem,
            image_context,
            canvas,
            texture_creator,
            event_pump,
            entities: vec![],
            components: vec![],
            systems: vec![],
            textures: vec![],
            currently_pressed_keys: HashSet::new(),
        })
    }

    pub fn run(&mut self) {
        'running: loop {
            for event in self.event_pump.poll_iter() {
                match event {
                    Event::Quit { .. }
                    | Event::KeyDown {
                        keycode: Some(Keycode::Escape),
                        ..
                    } => break 'running,
                    Event::KeyDown { keycode, .. } => {
                        self.currently_pressed_keys.insert(keycode.unwrap());
                    }
                    Event::KeyUp { keycode, .. } => {
                        self.currently_pressed_keys.remove(&keycode.unwrap());
                    }
                    _ => {}
                }
            }
            self.canvas.set_draw_color(Color::RGB(60, 180, 180));
            self.canvas.clear();
            for system in self.systems.clone() {
                let Err(err) = system.on_update(&mut self.context()) else {
                    continue;
                };
                println!("error occcurred updating system: {err}");
            }
            self.canvas.present();
            std::thread::sleep(Duration::new(0, 1_000_000_000u32 / 60))
        }
    }

    pub fn context<'context>(&'context mut self) -> Context<'context, 'game>
    where
        'game: 'context,
    {
        Context {
            id_counter: &mut self.id_counter,
            canvas: &mut self.canvas,
            texture_creator: &self.texture_creator,
            entities: &mut self.entities,
            systems: &mut self.systems,
            textures: &mut self.textures,
        }
    }
}
